/*******************************************************************************
    Copyright (c) 2014 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

#include "nvidia_uvm_common.h"
#include "nvidia_page_migration.h"
#include "nvidia_uvm_utils.h"
#include "nvidia_uvm_lite.h"
#include "nvidia_uvm_lite_counters.h"
#include "nvidia_uvm_channel_mgmt_common.h"

static struct kmem_cache *g_uvmChannelManagerCache  __read_mostly = NULL;
static struct kmem_cache *g_uvmRingbufferCache      __read_mostly = NULL;
static struct kmem_cache *g_uvmChannelCache         __read_mostly = NULL;
static struct kmem_cache *g_uvmPushbufferCache      __read_mostly = NULL;

static RM_STATUS _create_channel_pool(UvmGpuUuid *pGpuUuidStruct,
                                      UvmChannelPool *channelPool);
static void _destroy_channel_pool(UvmChannelPool *channelPool);

static RM_STATUS _create_channel_list(UvmChannelPool *channelPool,
                                      struct list_head *head,
                                      unsigned numChannels);
static void _destroy_channel_list(UvmChannelPool *channelPool,
                                  struct list_head *head);

static RM_STATUS _create_channel_resources(UvmChannelPool *channelPool,
                                           UvmChannel *channel);
static void _destroy_channel_resources(UvmChannelPool *channelPool,
                                       UvmChannel *channel);

static RM_STATUS _create_ringbuffer_pool(UvmGpuUuid *pGpuUuidStruct,
                                         UvmChannelPool *channelPool,
                                         UvmRingbufferPool *ringbufferPool);

static void _destroy_ringbuffer_pool(UvmChannelPool *channelPool,
                                     UvmRingbufferPool *ringbufferPool);

static RM_STATUS _create_ringbuffer_list(struct list_head *head,
                                         unsigned numRingbuffers,
                                         unsigned ringbufferSize,
                                         NvU64 cpuOffset,
                                         UvmGpuPointer gpuOffset);
static void _destroy_ringbuffer_list(struct list_head *head);

static void _destroy_pushbuffer_list(struct list_head *head);

static RM_STATUS _find_memory_in_ringbuffer(
        UvmRingbuffer *ringbuffer, NvU64 *offset);

static void _reclaim_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer);

static void _free_submitted_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer);

//
// Thread Safety:
//   Called during the startup of UVM.
//   This must be called and completed before any other operations in the API
//   can be used.
//
RM_STATUS uvm_initialize_channel_mgmt_api(void)
{
    UVM_DBG_PRINT("Entering\n");

    NV_KMEM_CACHE_CREATE(g_uvmChannelManagerCache,
                         "uvm_channel_manager_t",
                         UvmChannelManager);
    if (!g_uvmChannelManagerCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmRingbufferCache,
                         "uvm_ringbuffer_t",
                         UvmRingbuffer);
    if (!g_uvmRingbufferCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmChannelCache,
                         "uvm_channel_t",
                         UvmChannel);
    if (!g_uvmChannelCache)
        goto fail;

    NV_KMEM_CACHE_CREATE(g_uvmPushbufferCache,
                         "uvm_pushbuffer_t",
                         UvmPushbuffer);
    if (!g_uvmPushbufferCache)
        goto fail;

    return RM_OK;

fail:
    UVM_ERR_PRINT_RM_STATUS("Could not allocate channel mgmt API resources.",
                            RM_ERR_NO_MEMORY);

    kmem_cache_destroy_safe(&g_uvmPushbufferCache);
    kmem_cache_destroy_safe(&g_uvmChannelCache);
    kmem_cache_destroy_safe(&g_uvmRingbufferCache);
    kmem_cache_destroy_safe(&g_uvmChannelManagerCache);

    return RM_ERR_NO_MEMORY;
}

//
// Thread Safety:
//   Called during the shutdown of UVM.
//   This must be called and completed after all other operations in the API
//   are finished.
//
void uvm_deinitialize_channel_mgmt_api(void)
{
    UVM_DBG_PRINT("Entering\n");

    kmem_cache_destroy_safe(&g_uvmPushbufferCache);
    kmem_cache_destroy_safe(&g_uvmChannelCache);
    kmem_cache_destroy_safe(&g_uvmRingbufferCache);
    kmem_cache_destroy_safe(&g_uvmChannelManagerCache);
}

RM_STATUS uvm_create_channel_manager(UvmGpuUuid *gpuUuid,
                                     UvmChannelManager **channelManager)
{
    RM_STATUS status;
    UvmChannelManager *newChannelManager;

    UVM_DBG_PRINT_UUID("Entering", gpuUuid);

    newChannelManager = kmem_cache_zalloc(g_uvmChannelManagerCache,
                                          NV_UVM_GFP_FLAGS);
    if (!newChannelManager)
    {
        status = RM_ERR_NO_MEMORY;
        goto cleanup;
    }

    // Create channel pool
    status = _create_channel_pool(gpuUuid, &newChannelManager->channelPool);
    if (status != RM_OK)
        goto cleanup;

    // Create ringbuffer pool
    status = _create_ringbuffer_pool(gpuUuid,
                                    &newChannelManager->channelPool,
                                    &newChannelManager->ringbufferPool);
    if (status != RM_OK)
        goto cleanup;

    init_rwsem(&newChannelManager->channelManagerLock);

    *channelManager = newChannelManager;

    return RM_OK;

cleanup:

    if (newChannelManager && (newChannelManager->ringbufferPool.numRingbuffers))
        _destroy_ringbuffer_pool(&newChannelManager->channelPool,
                                 &newChannelManager->ringbufferPool);

    if (newChannelManager && (newChannelManager->channelPool.numChannels))
        _destroy_channel_pool(&newChannelManager->channelPool);

    return status;
}

void uvm_destroy_channel_manager(UvmChannelManager *channelManager)
{
    UVM_DBG_PRINT("Entering\n");

    _destroy_ringbuffer_pool(&channelManager->channelPool,
                             &channelManager->ringbufferPool);

    _destroy_channel_pool(&channelManager->channelPool);

    kmem_cache_free(g_uvmChannelManagerCache, channelManager);
}

//
// channelPool should be zeroed out before calling this function.
//
static RM_STATUS _create_channel_pool(UvmGpuUuid *pGpuUuidStruct,
                                      UvmChannelPool *channelPool)
{
    RM_STATUS status;
    unsigned numChannels = UVM_CHANNEL_POOL_DEFAULT_SIZE;
    unsigned long long uuidMsb = 0;
    unsigned long long uuidLsb = 0;
    struct list_head channelListHead;

    // Create a GPU Session
    status = nvUvmInterfaceSessionCreate(&channelPool->hSession);
    if (status != RM_OK)
    {
        UVM_ERR_PRINT_UUID("Could not create a session. "
                           "RM_STATUS: 0x%x.",
                           pGpuUuidStruct, status);
        goto cleanup;
    }

    // Create a VASpace
    memcpy(&uuidMsb, &pGpuUuidStruct->uuid[0], (UVM_UUID_LEN >> 1));
    memcpy(&uuidLsb, &pGpuUuidStruct->uuid[8], (UVM_UUID_LEN >> 1));
    status = nvUvmInterfaceAddressSpaceCreate(
            channelPool->hSession,
            uuidMsb,
            uuidLsb,
            &channelPool->hVaSpace);
    if (status != RM_OK)
    {
        UVM_ERR_PRINT_UUID("Could not create an address space. "
                           "RM_STATUS: 0x%x.",
                           pGpuUuidStruct, status);
        goto cleanup;
    }

    // Get GPU caps like ECC support on GPU, big page size, small page size, etc.
    status = nvUvmInterfaceQueryCaps(channelPool->hVaSpace,
                                     &channelPool->gpuCaps);
    if (status != RM_OK)
    {
        UVM_ERR_PRINT_UUID("Could not lookup GPU capabilities. "
                           "RM_STATUS: 0x%x.",
                           pGpuUuidStruct, status);
        goto cleanup;
    }

    // build channel list
    INIT_LIST_HEAD(&channelListHead);
    status = _create_channel_list(channelPool, &channelListHead, numChannels);
    if (status != RM_OK)
    {
        UVM_ERR_PRINT_UUID("Could not create channel list. "
                           "RM_STATUS: 0x%x.",
                           pGpuUuidStruct, status);
        goto cleanup;
    }

    INIT_LIST_HEAD(&channelPool->channelListHead);
    list_replace(&channelListHead, &channelPool->channelListHead);
    channelPool->numChannels = numChannels;

    return RM_OK;

cleanup:

    if (channelPool->hVaSpace)
    {
        nvUvmInterfaceAddressSpaceDestroy(channelPool->hVaSpace);
        channelPool->hVaSpace = NULL;
    }

    if (channelPool->hSession)
    {
        nvUvmInterfaceSessionDestroy(channelPool->hSession);
        channelPool->hSession = NULL;
    }

    return status;
}

static void _destroy_channel_pool(UvmChannelPool *channelPool)
{
    _destroy_channel_list(channelPool, &channelPool->channelListHead);

    nvUvmInterfaceAddressSpaceDestroy(channelPool->hVaSpace);
    nvUvmInterfaceSessionDestroy(channelPool->hSession);

    INIT_LIST_HEAD(&channelPool->channelListHead);
    channelPool->numChannels = 0;
    channelPool->hVaSpace    = NULL;
    channelPool->hSession    = NULL;
}

//
// ringbufferPool should be zeroed out before calling this function.
//
static RM_STATUS _create_ringbuffer_pool(UvmGpuUuid *pGpuUuidStruct,
                                         UvmChannelPool *channelPool,
                                         UvmRingbufferPool *ringbufferPool)
{
    RM_STATUS status;
    unsigned numRingbuffers = UVM_RINGBUFFER_POOL_DEFAULT_SIZE;
    unsigned ringbufferSize = UVM_RINGBUFFER_DEFAULT_SIZE;
    UvmGpuPointer gpuPtr = 0;
    void         *cpuPtr = NULL;
    struct list_head ringbufferListHead;

    // Allocate memory on sysmem and map to CPU.
    status = nvUvmInterfaceMemoryAllocSys(
            channelPool->hVaSpace,
            numRingbuffers * ringbufferSize,
            &gpuPtr);
    if (status != RM_OK)
    {
        UVM_ERR_PRINT("ERROR: could not allocate GPU memory"
                      " for ringbuffer pool. RM_STATUS: 0x%x.\n",
                      status);
        goto cleanup;
    }

    status = nvUvmInterfaceMemoryCpuMap(
            channelPool->hVaSpace,
            gpuPtr,
            numRingbuffers * ringbufferSize,
            &cpuPtr);
    if (status != RM_OK)
    {
        UVM_ERR_PRINT("ERROR: could not map GPU VA to CPU VA. "
                      "RM_STATUS: 0x%x.\n", status);
        goto cleanup;
    }

    // build ring buffer list
    INIT_LIST_HEAD(&ringbufferListHead);
    status = _create_ringbuffer_list(&ringbufferListHead,
                                     numRingbuffers,
                                     ringbufferSize,
                                     (NvU64)cpuPtr,
                                     gpuPtr);
    if (status != RM_OK)
        goto cleanup;

    INIT_LIST_HEAD(&ringbufferPool->ringbufferFreeListHead);
    list_replace(&ringbufferListHead,
                 &ringbufferPool->ringbufferFreeListHead);

    INIT_LIST_HEAD(&ringbufferPool->ringbufferBusyListHead);
    INIT_LIST_HEAD(&ringbufferPool->pbFreeListHead);

    ringbufferPool->numRingbuffers = numRingbuffers;
    ringbufferPool->gpuPtr         = gpuPtr;
    ringbufferPool->cpuPtr         = cpuPtr;

    return RM_OK;

cleanup:

    if (cpuPtr)
        nvUvmInterfaceMemoryCpuUnMap(channelPool->hVaSpace, cpuPtr);

    if (gpuPtr)
        nvUvmInterfaceMemoryFree(channelPool->hVaSpace, gpuPtr);

    return status;
}

static void _destroy_ringbuffer_pool(UvmChannelPool *channelPool,
                                     UvmRingbufferPool *ringbufferPool)
{
    // Free ringbuffer list
    _destroy_pushbuffer_list(&ringbufferPool->pbFreeListHead);
    _destroy_ringbuffer_list(&ringbufferPool->ringbufferFreeListHead);
    _destroy_ringbuffer_list(&ringbufferPool->ringbufferBusyListHead);

    // Unmap CPU mapping and free GPU memory allocation
    nvUvmInterfaceMemoryCpuUnMap(channelPool->hVaSpace,
                                 ringbufferPool->cpuPtr);
    nvUvmInterfaceMemoryFree(channelPool->hVaSpace,
                             ringbufferPool->gpuPtr);


    INIT_LIST_HEAD(&ringbufferPool->pbFreeListHead);
    INIT_LIST_HEAD(&ringbufferPool->ringbufferFreeListHead);
    INIT_LIST_HEAD(&ringbufferPool->ringbufferBusyListHead);
    ringbufferPool->numRingbuffers = 0;
    ringbufferPool->gpuPtr         = 0;
    ringbufferPool->cpuPtr         = 0;
}

static RM_STATUS _create_channel_list(UvmChannelPool *channelPool,
                                      struct list_head *head,
                                      unsigned numChannels)
{
    RM_STATUS status;
    unsigned i;
    UvmChannel *channel;

    for (i = 0; i < numChannels; ++i)
    {
        channel = kmem_cache_alloc(g_uvmChannelCache, NV_UVM_GFP_FLAGS);
        if (!channel)
        {
            status = RM_ERR_NO_MEMORY;
            UVM_ERR_PRINT("ERROR: could not allocate memory for channel "
                          "at index %d. RM_STATUS: 0x%x.\n", i, status);
            goto cleanup_channel_list;
        }

        status = _create_channel_resources(channelPool, channel);
        if (status != RM_OK)
        {
            UVM_ERR_PRINT("ERROR: could not create channel resources "
                          "at index %d. RM_STATUS: 0x%x.\n",
                          i, status);
            kmem_cache_free(g_uvmChannelCache, channel);
            goto cleanup_channel_list;
        }

        list_add(&channel->channelList, head);

        INIT_LIST_HEAD(&channel->pbSubmittedListHead);
    }

    return RM_OK;

cleanup_channel_list:
    _destroy_channel_list(channelPool, head);

    return status;
}

static void _destroy_channel_list(UvmChannelPool *channelPool,
                                  struct list_head *head)
{
    if (!list_empty(head))
    {
        struct list_head *pos;
        struct list_head *safepos;
        UvmChannel       *channel;

        list_for_each_safe(pos, safepos, head)
        {
            channel = list_entry(pos, UvmChannel, channelList);
            list_del(pos);
            _destroy_channel_resources(channelPool, channel);
            kmem_cache_free(g_uvmChannelCache, channel);
        }
    }
}

static RM_STATUS _create_channel_resources(UvmChannelPool *channelPool,
                                           UvmChannel *channel)
{
    RM_STATUS status;
    unsigned ceInstance;
    UvmGpuPointer semaGpuPointer;
    void         *semaCpuPointer;
    const unsigned SEMA_SIZE = 0x1000;

    channel->trackingInfo.semaGpuPointer = 0;
    channel->trackingInfo.semaCpuPointer = NULL;

    // Get all the channel pointers
    status = nvUvmInterfaceChannelAllocate(
            channelPool->hVaSpace,
            &channel->hChannel,
            &channel->channelInfo);
    if (status != RM_OK)
        return status;

    // Allocate copy engine object
    for (ceInstance = 1; ceInstance <= MAX_NUM_COPY_ENGINES; ++ceInstance)
    {
        status = nvUvmInterfaceCopyEngineAllocate(channel->hChannel,
                                                  ceInstance,
                                                  &channel->ceClassNumber,
                                                  &channel->hCopyEngine);

        if ((status == RM_ERR_INVALID_INDEX) || (status == RM_OK))
            break;
    }
    if (status != RM_OK)
        goto cleanup_channel;

    // Set up CE hal functions
    status = NvUvmHalInit(
            channel->ceClassNumber,
            channel->channelInfo.channelClassNum,
            &channel->ceOps);
    if (RM_OK != status)
        goto cleanup_channel;

    status = nvUvmInterfaceMemoryAllocSys(channelPool->hVaSpace,
                                          SEMA_SIZE,
                                         &semaGpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not allocate GPU memory for sema.",
                                status);
        goto cleanup_channel;
    }

    status = nvUvmInterfaceMemoryCpuMap(channelPool->hVaSpace,
                                        semaGpuPointer,
                                        SEMA_SIZE, &semaCpuPointer);
    if (RM_OK != status)
    {
        UVM_ERR_PRINT_RM_STATUS("Could not map sema to CPU VA.", status);
        goto cleanup_channel;
    }

    // zero the semaphore region
    memset(semaCpuPointer, 0x0, SEMA_SIZE);
    
    // TODO: this is to help testing GPFIFO reclaim.
    // REMOVE it later.
    channel->channelInfo.numGpFifoEntries = 5;

    // Initialize channel tracking info
    channel->trackingInfo.curGpFifoOffset          = 0;
    channel->trackingInfo.numReservedGpFifoEntries = 0;
    channel->trackingInfo.numFreeGpFifoEntries =
        channel->channelInfo.numGpFifoEntries;

    channel->trackingInfo.semaGpuPointer = semaGpuPointer;
    channel->trackingInfo.semaCpuPointer = semaCpuPointer;
    channel->trackingInfo.numCompletions = 0;
    channel->trackingInfo.numSubmissions = 0;

    return RM_OK;

cleanup_channel:
    _destroy_channel_resources(channelPool, channel);

    return status;
}

static void _destroy_channel_resources(
        UvmChannelPool *channelPool, UvmChannel *channel)
{
    if (channel->trackingInfo.semaCpuPointer)
        nvUvmInterfaceMemoryCpuUnMap(
                channelPool->hVaSpace,
                channel->trackingInfo.semaCpuPointer);

    if (channel->trackingInfo.semaGpuPointer)
        nvUvmInterfaceMemoryFree(
                channelPool->hVaSpace,
                channel->trackingInfo.semaGpuPointer);

    nvUvmInterfaceChannelDestroy(channel->hChannel);
}

static RM_STATUS _create_ringbuffer_list(struct list_head *head,
                                         unsigned numRingbuffers,
                                         unsigned ringbufferSize,
                                         NvU64 cpuOffset,
                                         UvmGpuPointer gpuOffset)
{
    unsigned i;
    UvmRingbuffer *ringbuffer;

    for (i = 0; i < numRingbuffers; ++i)
    {
        ringbuffer = kmem_cache_alloc(g_uvmRingbufferCache, NV_UVM_GFP_FLAGS);
        if (!ringbuffer)
        {
            UVM_ERR_PRINT("ERROR: could not allocate memory for ringbuffer "
                          "at index %d.\n", i);
            goto cleanup_ringbuffer_list;
        }

        //
        // Initalize ringbuffer
        //

        ringbuffer->cpuBegin  = (void*)cpuOffset;
        ringbuffer->cpuEnd    = (void*)(cpuOffset + ringbufferSize);
        ringbuffer->gpuBegin  = gpuOffset;
        ringbuffer->curOffset = 0;

        INIT_LIST_HEAD(&ringbuffer->pbListHead);

        list_add(&ringbuffer->ringbufferList, head);

        cpuOffset += ringbufferSize;
        gpuOffset += ringbufferSize;
    }

    return RM_OK;

cleanup_ringbuffer_list:
    _destroy_ringbuffer_list(head);

    return RM_ERR_NO_MEMORY;
}

static void _destroy_ringbuffer_list(struct list_head *head)
{
    if (!list_empty(head))
    {
        struct list_head *pos;
        struct list_head *safepos;
        UvmRingbuffer    *ringbuffer;

        list_for_each_safe(pos, safepos, head) 
        {
            ringbuffer = list_entry(pos, UvmRingbuffer, ringbufferList);
            list_del(pos);

            _destroy_pushbuffer_list(&ringbuffer->pbListHead);

            kmem_cache_free(g_uvmRingbufferCache, ringbuffer);
        }
    }
}

static void _destroy_pushbuffer_list(struct list_head *head)
{
    if (!list_empty(head))
    {
        struct list_head *pos;
        struct list_head *safepos;
        UvmPushbuffer    *pushbuffer;

        list_for_each_safe(pos, safepos, head) 
        {
            pushbuffer = list_entry(pos, UvmPushbuffer, pbList);
            list_del(pos);

            kmem_cache_free(g_uvmPushbufferCache, pushbuffer);
        }
    }
}

void uvm_lock_channel_manager(UvmChannelManager *channelManager)
{
    down_write(&channelManager->channelManagerLock);
}

void uvm_unlock_channel_manager(UvmChannelManager *channelManager)
{
    up_write(&channelManager->channelManagerLock);
}

RM_STATUS uvm_channel_manager_alloc_pushbuffer_structure(
        UvmChannelManager *channelManager, UvmPushbuffer **pushbuffer)
{
    UvmPushbuffer *newPushbuffer;
    struct list_head *entry;
    struct list_head *safepos;

    // Find the first free pb structure in the ringbuffer pool.
    list_for_each_safe(
            entry, safepos, &channelManager->ringbufferPool.pbFreeListHead)
    {
        newPushbuffer = list_entry(entry, UvmPushbuffer, pbList);
        list_del(entry);

        goto done;
    }

    // Allocate pushbuffer structure, since we couldn't find any free pb.
    newPushbuffer = kmem_cache_alloc(g_uvmPushbufferCache, NV_UVM_GFP_FLAGS);
    if (!newPushbuffer)
        return RM_ERR_NO_MEMORY;

done:
    *pushbuffer = newPushbuffer;

    return RM_OK;
}

void uvm_channel_manager_free_pushbuffer_structure(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer)
{
    //
    // TODO: Use some heruistics to determine whether or not we need
    //       to release some memory back to the OS.
    //
    // Right now, we're just putting it back on the free list.
    //

    list_move(
            &pushbuffer->pbList,
            &channelManager->ringbufferPool.pbFreeListHead);
}

RM_STATUS uvm_channel_manager_get_ringbuffer(
        UvmChannelManager *channelManager, UvmRingbuffer **ringbuffer)
{
    struct list_head *entry;
    struct list_head *safepos;
    UvmRingbuffer *freeRingbuffer;
    UvmRingbufferPool *ringbufferPool = &channelManager->ringbufferPool;

    //
    // Find the first available ringbuffer with enough space in the pool.
    // Move it to busy list and return it to the caller.
    //
    list_for_each_safe(
            entry, safepos, &ringbufferPool->ringbufferFreeListHead)
    {
        NvU64 curAddr;

        freeRingbuffer = list_entry(entry, UvmRingbuffer, ringbufferList);

        if (RM_OK == _find_memory_in_ringbuffer(freeRingbuffer, &curAddr))
        {
            NvU64 cpuBegin = (NvU64)freeRingbuffer->cpuBegin;

            freeRingbuffer->curOffset = curAddr - cpuBegin;
            list_move_tail(entry, &ringbufferPool->ringbufferBusyListHead);

            *ringbuffer = freeRingbuffer;
            return RM_OK;
        }
    }

    return RM_ERR_NO_MEMORY;
}

//
// Find the immediate next slot that fits a pushbuffer.
//
static RM_STATUS _find_memory_in_ringbuffer(
        UvmRingbuffer *ringbuffer, NvU64 *offset)
{
    UvmPushbuffer *pbHead;
    UvmPushbuffer *pbTail;
    NvU64 pbHeadBegin;
    NvU64 pbHeadEnd;
    NvU64 pbTailBegin;
    NvU64 pbTailEnd;
    NvU64 ringbufferBegin;
    NvU64 ringbufferEnd;

    ringbufferBegin = (NvU64)ringbuffer->cpuBegin;
    ringbufferEnd   = (NvU64)ringbuffer->cpuEnd;

    // Empty list, we start from the beginning.
    if (list_empty(&ringbuffer->pbListHead))
    {
        *offset = ringbufferBegin;
        return RM_OK;
    }

    //
    // Single push buffer in the list, we look for space in
    // top & bottom of ringbuffer.
    //
    // list_is_singular not implemented on all Linux versions.
    // Consider replacing below with list_is_singular in the future.
    //
    if (!list_empty(&ringbuffer->pbListHead) &&
        (ringbuffer->pbListHead.next == ringbuffer->pbListHead.prev))
    {
        pbHead = 
            list_entry(
                    ringbuffer->pbListHead.next,
                    UvmPushbuffer,
                    pbList);
        pbHeadBegin = (NvU64)pbHead->cpuBegin;
        pbHeadEnd   = pbHeadBegin + pbHead->curOffset;

        if (ringbufferEnd - pbHeadEnd >= UVM_PUSHBUFFER_RESERVATION_SIZE)
        {
            *offset = pbHeadEnd;
            return RM_OK;
        }
        else if (ringbufferBegin - pbHeadBegin >= UVM_PUSHBUFFER_RESERVATION_SIZE)
        {
            *offset = ringbufferBegin;
            return RM_OK;
        }
        else
        {
            return RM_ERR_NO_MEMORY;
        }
    }

    //
    // A list of multiple pushbuffers, we look for space in:
    // 1. top of ringbuffer
    // 2. bottom of ringbuffer
    // 3. between pb head and tail
    //

    pbHead = 
        list_entry(
                ringbuffer->pbListHead.next,
                UvmPushbuffer,
                pbList);
    pbHeadBegin = (NvU64)pbHead->cpuBegin;
    pbHeadEnd   = pbHeadBegin + pbHead->curOffset;

    pbTail =
        list_entry(
                ringbuffer->pbListHead.prev,
                UvmPushbuffer,
                pbList);
    pbTailBegin = (NvU64)pbTail->cpuBegin;
    pbTailEnd   = pbTailBegin + pbTail->curOffset;

    if (pbHeadBegin > pbTailBegin)
    {
        if (pbHeadBegin - pbTailEnd >= UVM_PUSHBUFFER_RESERVATION_SIZE)
        {
            *offset = pbTailEnd;
            return RM_OK;
        }
        else
        {
            return RM_ERR_NO_MEMORY;
        }
    }
    else
    {
        if (ringbufferEnd - pbTailEnd >= UVM_PUSHBUFFER_RESERVATION_SIZE)
        {
            *offset = pbTailEnd;
            return RM_OK;
        }
        else if (pbHeadBegin - ringbufferBegin >= UVM_PUSHBUFFER_RESERVATION_SIZE)
        {
            *offset = ringbufferBegin;
            return RM_OK;
        }
        else
        {
            return RM_ERR_NO_MEMORY;
        }
    }

    return RM_ERR_NO_MEMORY;
}

void uvm_channel_manager_put_ringbuffer(
        UvmChannelManager *channelManager, UvmRingbuffer *ringbuffer)
{
    //
    // Put it the tail of the free list.
    // This way, the ringbuffer in the front get more time to free up space.
    //
    list_move_tail(
            &ringbuffer->ringbufferList,
            &channelManager->ringbufferPool.ringbufferFreeListHead);
}

void uvm_channel_manager_associate_pushbuffer_to_ringbuffer(
        UvmRingbuffer *ringbuffer, UvmPushbuffer *pushbuffer)
{
    list_add_tail(&pushbuffer->pbList, &ringbuffer->pbListHead);
}

RM_STATUS uvm_channel_manager_get_channel(
        UvmChannelManager *channelManager, UvmChannel **channel)
{
    struct list_head *entry;
    struct list_head *safepos;
    UvmChannel *freeChannel;
    UvmChannelPool *channelPool = &channelManager->channelPool;

    //
    // Find the first available channel with spare GPFIFO entries.
    // Use a round robin policy.
    // Also reserve a GPFIFO entry.
    //
    list_for_each_safe(
            entry, safepos, &channelPool->channelListHead)
    {
        freeChannel = list_entry(entry, UvmChannel, channelList);

        if (freeChannel->trackingInfo.numFreeGpFifoEntries)
        {
            list_move_tail(entry, &channelPool->channelListHead);

            --freeChannel->trackingInfo.numFreeGpFifoEntries;
            ++freeChannel->trackingInfo.numReservedGpFifoEntries;

            *channel = freeChannel;

            return RM_OK;
        }
    }

    return RM_ERR_INSUFFICIENT_RESOURCES;
}

void uvm_channel_manager_put_channel(
        UvmChannelManager *channelManager, UvmChannel *channel)
{
    //
    // Put the channel back the to head of the channel pool.
    // And un-reserve the GPFIFO entry.
    //
    // This makes the channel ready for submitting a pushbuffer right away.
    //

    list_move(
            &channel->channelList, 
            &channelManager->channelPool.channelListHead);

    --channel->trackingInfo.numReservedGpFifoEntries;
    ++channel->trackingInfo.numFreeGpFifoEntries;
}

void uvm_channel_manager_associate_pushbuffer_to_channel(
        UvmChannel *channel, UvmPushbuffer *pushbuffer)
{
    list_add_tail(&pushbuffer->pbSubmittedList, &channel->pbSubmittedListHead);

    --channel->trackingInfo.numReservedGpFifoEntries;
    ++channel->trackingInfo.numSubmissions;

    uvm_channel_manager_update_channel_progress(channel);
}

//
// The general idea of reclaim is:
//
// 1. We look at the head of each ringbuffer.
// 2. For each completed head, we look into the channel it's associated with.
//    * We free all the pushbuffers that are completed before this head in
//      that channel.
//    * This way, we should get to free a bunch of GPFIFO entries as well.
//
RM_STATUS uvm_channel_manager_reclaim(UvmChannelManager *channelManager)
{
    UvmRingbufferPool  *ringbufferPool;
    UvmRingbuffer      *ringbuffer;
    UvmPushbuffer      *pushbuffer;
    UvmChannelTracking *trackingInfo;
    NvU32 newSema;
    NvU32 oldSema;

    ringbufferPool = &channelManager->ringbufferPool;

    UVM_DBG_PRINT("Reclaim is triggered.\n");

    //
    // Check the head of each ringbuffer to find out if they've completed.
    //
    list_for_each_entry(
            ringbuffer,
            &ringbufferPool->ringbufferFreeListHead,
            ringbufferList)
    {
        if (!list_empty(&ringbuffer->pbListHead))
        {
            pushbuffer =
                list_entry(
                        ringbuffer->pbListHead.next, UvmPushbuffer, pbList);

            trackingInfo = &pushbuffer->channel->trackingInfo;

            newSema = UVM_READ_SEMA(trackingInfo->semaCpuPointer);
            oldSema = (NvU32)trackingInfo->numCompletions;

            // Check if the pb is completed.
            if (newSema >= (NvU32)pushbuffer->seqNum ||
                newSema < oldSema)
            {
                _reclaim_pushbuffer(channelManager, pushbuffer);
            }
        }
    }

    return RM_OK;
}

//
// Remove all the completed PBs in the same channel.
// Update GPFIFO queue book-keeping.
//
static void _reclaim_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer)
{
    NvU32 curCompletedGpFifoOffset;
    NvU32 curSubmittedGpFifoOffset;
    UvmPushbuffer *pbEntry;
    UvmPushbuffer *pbSafepos;

    curCompletedGpFifoOffset = pushbuffer->gpFifoOffset;
    curSubmittedGpFifoOffset = pushbuffer->channel->trackingInfo.curGpFifoOffset;

    //
    // C: curCompletedGpFifoOffset
    // S: curSubmittedGpFifoOffset
    // -: currently pending completion
    // =: already completed
    //
    // [-----S==========C---------]
    if (curCompletedGpFifoOffset > curSubmittedGpFifoOffset)
    {
        list_for_each_entry_safe(
                pbEntry, 
                pbSafepos,
                &pushbuffer->channel->pbSubmittedListHead,
                pbSubmittedList)
        {
            if (pbEntry->gpFifoOffset >= curSubmittedGpFifoOffset &&
                pbEntry->gpFifoOffset <= curCompletedGpFifoOffset)
            {
                _free_submitted_pushbuffer(channelManager, pbEntry);
            }
        }
    }
    // [=====C----------S=========]
    else
    {
        list_for_each_entry_safe(
                pbEntry, 
                pbSafepos,
                &pushbuffer->channel->pbSubmittedListHead,
                pbSubmittedList)
        {
            if (pbEntry->gpFifoOffset > curSubmittedGpFifoOffset ||
                pbEntry->gpFifoOffset <= curCompletedGpFifoOffset)
            {
                _free_submitted_pushbuffer(channelManager, pbEntry);
            }
        }
    }
}

//
// Free memory associated of this pushbuffer back into ringbuffer,
// and put the pb data structure back into the free list.
//
static void _free_submitted_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer)
{
    UVM_DBG_PRINT(
            "Free Pushbuffer: { begin: 0x%llx, length: 0x%llx }\n",
            (NvU64)pushbuffer->cpuBegin,
            pushbuffer->curOffset);

    ++pushbuffer->channel->trackingInfo.numFreeGpFifoEntries;

    list_del(&pushbuffer->pbSubmittedList);

    uvm_channel_manager_free_pushbuffer_structure(
            channelManager, pushbuffer);
}

void uvm_channel_manager_update_all_channel_progress(
        UvmChannelManager *channelManager)
{
    UvmChannel *channel;
    
    list_for_each_entry(
            channel,
            &channelManager->channelPool.channelListHead,
            channelList)
    {
        uvm_channel_manager_update_channel_progress(channel);
    }
}
