/*******************************************************************************
    Copyright (c) 2014 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

#include "nvidia_uvm_channel_mgmt_common.h"

// Lock channel manager before calling this function.
RM_STATUS uvm_get_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer **pushbuffer)
{
    RM_STATUS status;
    UvmRingbuffer *ringbuffer    = NULL;
    UvmChannel    *channel       = NULL;
    UvmPushbuffer *newPushbuffer = NULL;

    uvm_lock_channel_manager(channelManager);

    // Look for avalable ringbuffer.
    while (RM_OK !=
            uvm_channel_manager_get_ringbuffer(channelManager, &ringbuffer))
    {
        uvm_channel_manager_update_all_channel_progress(channelManager);
        uvm_channel_manager_reclaim(channelManager);

        uvm_unlock_channel_manager(channelManager);
        uvm_lock_channel_manager(channelManager);
    }

    // Get a channel with spare GPFIFO entries.
    while (RM_OK !=
            uvm_channel_manager_get_channel(channelManager, &channel))
    {
        uvm_channel_manager_update_all_channel_progress(channelManager);
        uvm_channel_manager_reclaim(channelManager);

        uvm_unlock_channel_manager(channelManager);
        uvm_lock_channel_manager(channelManager);
    }

    // Allocate a pushbuffer structure.
    status =
        uvm_channel_manager_alloc_pushbuffer_structure(
                channelManager, &newPushbuffer);
    if (status != RM_OK)
        goto cleanup;

    // Associate the pushbuffer with the ringbuffer.
    uvm_channel_manager_associate_pushbuffer_to_ringbuffer(
            ringbuffer, newPushbuffer);

    // Initialize pushbuffer structure.
    newPushbuffer->cpuBegin   = ringbuffer->cpuBegin + ringbuffer->curOffset;
    newPushbuffer->gpuBegin   = ringbuffer->gpuBegin + ringbuffer->curOffset;
    newPushbuffer->pbOffset   = newPushbuffer->cpuBegin;
    newPushbuffer->curOffset  = 0;
    newPushbuffer->ringbuffer = ringbuffer;
    newPushbuffer->channel    = channel;

    //
    // Reserve space for the initial semaphore acquire by
    //  1. bumping up the pbOffset
    //  2. bumping up the curOffset
    //
    newPushbuffer->curOffset +=
        channel->ceOps.semaphoreAcquire_GEQ(
                (unsigned**)&newPushbuffer->pbOffset,
                (unsigned*)((NvU64)newPushbuffer->cpuBegin +
                    UVM_PUSHBUFFER_RESERVATION_SIZE),
                channel->trackingInfo.semaGpuPointer,
                0);

    *pushbuffer = newPushbuffer;

    uvm_unlock_channel_manager(channelManager);

    return RM_OK;

cleanup:

    if (newPushbuffer)
        uvm_channel_manager_free_pushbuffer_structure(
                channelManager, newPushbuffer);

    if (channel)
        uvm_channel_manager_put_channel(channelManager, channel);
    
    if (ringbuffer)
        uvm_channel_manager_put_ringbuffer(channelManager, ringbuffer);

    uvm_unlock_channel_manager(channelManager);

    return status;
}

// Lock channel manager before calling this function.
void uvm_cancel_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer)
{
    UvmChannel *channel;
    UvmRingbuffer *ringbuffer;

    uvm_lock_channel_manager(channelManager);

    channel    = pushbuffer->channel;
    ringbuffer = pushbuffer->ringbuffer;

    uvm_channel_manager_free_pushbuffer_structure(
            channelManager, pushbuffer);

    uvm_channel_manager_put_channel(channelManager, channel);

    uvm_channel_manager_put_ringbuffer(channelManager, ringbuffer);

    uvm_unlock_channel_manager(channelManager);
}

// Lock channel manager before calling this function.
RM_STATUS uvm_submit_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer,
        UvmTracker *trackerToWait, UvmTracker *newTracker)
{
    RM_STATUS status;
    NvU64 numMethods = 0;
    UvmChannel *channel;
    UvmRingbuffer *ringbuffer;

    uvm_lock_channel_manager(channelManager);

    channel    = pushbuffer->channel;
    ringbuffer = pushbuffer->ringbuffer;

    if (!channel->trackingInfo.numReservedGpFifoEntries)
    {
        status = RM_ERR_INSUFFICIENT_RESOURCES;
        goto cleanup;
    }

    // Push acquire value in the very beginning of the pushbuffer.
    if (trackerToWait)
    {
        void *cpuBegin = pushbuffer->cpuBegin;

        //
        // TODO: determine whether or not we can avoid this acquire.
        //

        numMethods = channel->ceOps.semaphoreAcquire_GEQ(
                (unsigned **)&cpuBegin,
                (unsigned*)((NvU64)pushbuffer->cpuBegin +
                            UVM_PUSHBUFFER_RESERVATION_SIZE),
                trackerToWait->item.channel->trackingInfo.semaGpuPointer,
                trackerToWait->item.seqNum);
        if (!numMethods)
        {
            status = RM_ERR_NO_MEMORY;
            goto cleanup;
        }
    }

    // Push release value.
    numMethods = channel->ceOps.semaphoreRelease(
            (unsigned **)&pushbuffer->pbOffset,
            (unsigned*)((NvU64)pushbuffer->cpuBegin +
                        UVM_PUSHBUFFER_RESERVATION_SIZE),
            channel->trackingInfo.semaGpuPointer,
            channel->trackingInfo.numSubmissions + 1);
    if (!numMethods)
    {
        status = RM_ERR_NO_MEMORY;
        goto cleanup;
    }
    pushbuffer->curOffset += numMethods;

    // Write the GP entry.
    channel->ceOps.writeGpEntry(
            channel->channelInfo.gpFifoEntries,
            channel->trackingInfo.curGpFifoOffset,
            pushbuffer->gpuBegin,
            pushbuffer->curOffset);

    // Increment current GPFIFO offset and launch pushbuffer
    if (++channel->trackingInfo.curGpFifoOffset >=
        channel->channelInfo.numGpFifoEntries)
    {
        channel->trackingInfo.curGpFifoOffset = 0;
    }
    NvUvmChannelWriteGpPut(channel->channelInfo.GPPut,
                           channel->trackingInfo.curGpFifoOffset);

    // Submit to channel book-keeping.
    uvm_channel_manager_associate_pushbuffer_to_channel(channel, pushbuffer);
    pushbuffer->gpFifoOffset = channel->trackingInfo.curGpFifoOffset;
    pushbuffer->seqNum       = channel->trackingInfo.numSubmissions;
    if (newTracker)
    {
        newTracker->item.seqNum  = channel->trackingInfo.numSubmissions;
        newTracker->item.channel = channel;
    }

    // Put ringbuffer back to free list.
    uvm_channel_manager_put_ringbuffer(channelManager, ringbuffer);

    uvm_unlock_channel_manager(channelManager);

    return RM_OK;

cleanup:

    uvm_unlock_channel_manager(channelManager);

    return status;
}

// No lock required for this call.
RM_STATUS uvm_wait_for_tracker(UvmTracker *tracker)
{
    UvmChannelTracking *trackingInfo = &tracker->item.channel->trackingInfo;
    NvU32 seqNum_32 = (NvU32)tracker->item.seqNum;

    while(1)
    {
        //
        // This code assumes that there will be fewer than 2^32 pushbuffers
        // completed since last time we submitted a pushbuffer to this channel.
        //
        // Which is always guaranteed, since each channel allows up to only 2k
        // pushbuffers pending completion.
        //

        NvU32 newSema = UVM_READ_SEMA(trackingInfo->semaCpuPointer);
        NvU32 oldSema = (NvU32)trackingInfo->numCompletions;

        if (newSema >= seqNum_32)
            break;
        else if (newSema < oldSema)     // wrap-around happened.
            break;
    }

    return RM_OK;
}

void uvm_channel_manager_update_channel_progress(UvmChannel *channel)
{
    UvmChannelTracking *trackingInfo;
    NvU64 numCompletions;
    NvU32 oldSema;
    NvU32 newSema;

    trackingInfo = &channel->trackingInfo;

    numCompletions = trackingInfo->numCompletions;
    oldSema = (NvU32)numCompletions;
    newSema = UVM_READ_SEMA(trackingInfo->semaCpuPointer);

    if (oldSema != newSema)
    {
        numCompletions &= 0xFFFFFFFF00000000ull;

        if (newSema < oldSema)
            numCompletions += (1ull << 32);

        numCompletions |= (NvU64)newSema;

        trackingInfo->numCompletions = numCompletions;
    }
}
